from pycocotools.coco import COCO
import numpy as np
import random
import os
import cv2

import sys
sys.path.append('./')

### For visualizing the outputs ###
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import torch
from torch.utils.data import Dataset, DataLoader
from torchvision import io, transforms, utils
import torchvision.transforms.functional as TF
from tqdm.auto import tqdm
from torchvision.io.image import read_image
from torchvision.models.segmentation import DeepLabV3_ResNet101_Weights, deeplabv3_resnet101, lraspp_mobilenet_v3_large, LRASPP_MobileNet_V3_Large_Weights
from torchvision.transforms.functional import to_pil_image
from pathlib import Path
import random
from typing import Any, Callable, List, Optional, Tuple
import torch.optim as optim
from torch import nn, einsum
from torch.autograd import Variable
import torchgeometry as tgm

import torch
import torch.utils.data

from Guided.dataset.helpers import get_splitted_dataset
from Guided.helpers import get_parser, Operation, OptimizerDetails
from Guided.models.resnet import ResNet18_64x64, ResNet18_64x64_1, ResNet18_256x256
from scripts.imagenet import get_loader_from_dataset, get_train_val_datasets
import cv2

import torchvision
import cv2
from torchvision import transforms, utils
import torchvision.models as models
from torch.utils import data
import torch.nn.functional as F
import os
import errno
import shutil

def create_folder(path):
    try:
        os.mkdir(path)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise
        pass

torch.manual_seed(0)

parser = get_parser()
parser.add_argument('--root', default='ICML')
parser.add_argument("--lr", default=1e-4, type=float)
parser.add_argument("--momentum", default=0.9, type=float)
parser.add_argument("--wd", default=1e-2, type=float)
parser.add_argument("--shuffle", default=False, help='shuffles the data when we can the train and val data')
parser.add_argument('--direct', action='store_true', help='use direct sampling for noising and denoising')
parser.add_argument('--epochs', default=100, type=int)
parser.add_argument('--run_command', default='PYTHONPATH=. python Guided/membership_classification.py',
                    help='How to run the script.')
parser.add_argument('--save_every', type=int, default=1)
parser.add_argument('--test_every', type=int, default=10)
parser.add_argument('--optimizer', default='adamw', choices=['sgd', 'adamw'])
parser.add_argument('--use_noise', action='store_true')
parser.add_argument('--fixed_noise', action='store_true')
parser.add_argument('--almost_fixed_noise', action='store_true')
parser.add_argument('--distribution', action='store_true')
parser.add_argument('--load', action='store_true')
parser.add_argument('--remove_bn', action='store_true', default=False)
parser.add_argument('--repeat', type=int, default=1)
parser.add_argument('--use_image', type=int, default=1)
parser.add_argument('--wandb', type=int, default=1)
parser.add_argument('--input_size', type=int, default=64)

parser.add_argument("--optim_lr", default=1e-2, type=float)
parser.add_argument('--optim_max_iters', type=int, default=1)
parser.add_argument('--optim_mask_type', type=int, default=1)
parser.add_argument("--optim_loss_cutoff", default=0.00001, type=float)
parser.add_argument('--optim_guidance_3', action='store_true', default=False)
parser.add_argument('--optim_original_guidance', action='store_true', default=False)
parser.add_argument("--optim_guidance_3_wt", default=2.0, type=float)
parser.add_argument('--optim_do_guidance_3_norm', action='store_true', default=False)
parser.add_argument('--optim_unscaled_guidance_3', action='store_true', default=False)
parser.add_argument("--optim_tv_loss", default=None, type=float)
parser.add_argument('--optim_warm_start', action='store_true', default=False)
parser.add_argument('--optim_print', action='store_true', default=False)
parser.add_argument('--optim_aug', action='store_true', default=False)
parser.add_argument('--optim_folder', default='./temp/')
parser.add_argument('--optim_sampling_type', default=None)
parser.add_argument("--optim_num_steps", nargs="+", default=[1], type=int)
parser.add_argument("--optim_mask_fraction", default=0.5, type=float)


args = parser.parse_args()

# MODEL_FLAGS="--attention_resolutions 32,16,8 --class_cond False --diffusion_steps 1000 --image_size 256 --learn_sigma True --noise_schedule linear --num_channels 256 --num_head_channels 64 --num_res_blocks 2 --resblock_updown True --use_fp16 False --use_scale_shift_norm True"
# PYTHONPATH=. python scripts/load_model.py $MODEL_FLAGS --classifier_scale 10.0 --classifier_path models/256x256_classifier.pt --model_path models/256x256_diffusion_uncond.pt $SAMPLE_FLAGS

# MODEL_FLAGS="--attention_resolutions 32,16,8 --class_cond False --diffusion_steps 1000 --image_size 256 --learn_sigma True --noise_schedule linear --num_channels 256 --num_head_channels 64 --num_res_blocks 2 --resblock_updown True --use_fp16 False --use_scale_shift_norm True"
# PYTHONPATH=. python Guided/Segmentation_mobilenet.py $MODEL_FLAGS --classifier_scale 10.0 --classifier_path models/256x256_classifier.pt --model_path models/256x256_diffusion_uncond.pt $SAMPLE_FLAGS --batch_size 4



BATCH_SIZE = args.batch_size

model = models.resnet18(pretrained=True)
Trans = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

class Complete_Model(nn.Module):
    def __init__(self, model, Trans):
        super(Complete_Model, self).__init__()
        self.trans = Trans
        self.model = model

    def forward(self, x):
        x = (x + 1) * 0.5
        x = TF.resize(x, (224, 224), interpolation=TF.InterpolationMode.BILINEAR)
        x = self.trans(x)
        x = self.model(x)
        return x

com_model = Complete_Model(model, Trans).cuda()

class Grad:
    def __init__(self, model):
        self.model = model

    def set_label(self, label):
        self.label = label

    def get_grads(self, image):
        All_Grads = []
        for i in range(image.shape[0]):
            im = image[i:i+1]
            tr = self.label[i:i+1]

            prediction = self.model(im)
            c = nn.CrossEntropyLoss(reduce=False)
            l = c(prediction, tr)

            self.model.zero_grad()
            input_gradient = torch.autograd.grad(l, self.model.parameters())
            input_gradient = [grad.detach() for grad in input_gradient]

            All_Grads.append(input_gradient)

        return All_Grads
        

operation_grad = Grad(com_model)

def mse_loss(input, target):
    all_loss = []
    for i in range(len(input)):
        inp = input[i]
        tar = target[i]

        l=0
        for j in range(len(inp)):
            l += ((inp[j] - tar[j]) ** 2).mean()

        all_loss.append(l)

    all_loss = torch.stack(all_loss)
    return all_loss




results_folder = args.optim_folder
create_folder(results_folder)

operation = OptimizerDetails()

operation.num_steps = args.optim_num_steps #[2]

operation.optimizer = 'Adam'
operation.lr = args.optim_lr
operation.loss_func = mse_loss

operation.max_iters = args.optim_max_iters
operation.loss_cutoff = args.optim_loss_cutoff
operation.tv_loss = args.optim_tv_loss
operation.operation_func = None


operation.guidance_3 = args.optim_guidance_3
operation.original_guidance = args.optim_original_guidance
operation.mask_type = args.optim_mask_type

operation.optim_guidance_3_wt = args.optim_guidance_3_wt
operation.do_guidance_3_norm = args.optim_do_guidance_3_norm
operation.optim_unscaled_guidance_3 = args.optim_unscaled_guidance_3
operation.sampling_type = args.optim_sampling_type

operation.warm_start = args.optim_warm_start
operation.print = args.optim_print
operation.print_every = 1
operation.folder = results_folder


# operation = [2, operation_func, optim.Adam, 0.001, nn.MSELoss(), 1000, 0.001]

operator = Operation(args, operation=operation, shape=[BATCH_SIZE, 3, 256, 256], progressive=True)
cnt = 0

def return_cv2(img, path):
    black = [255, 255, 255]
    img = (img + 1) * 0.5
    utils.save_image(img, path, nrow=1)
    img = cv2.imread(path)
    img = cv2.copyMakeBorder(img, 10, 10, 10, 10, cv2.BORDER_CONSTANT, value=black)
    return img

print('loading the dataset...')
train_dataset, val_dataset = get_train_val_datasets(args)
print('done')
print('splitting the dataset...')
train1, train2 = get_splitted_dataset(dataset=train_dataset,
                                      checkpoint_path='checkpoints/non_equal_split/partitions_train.pt')
val1, val2 = get_splitted_dataset(dataset=val_dataset, checkpoint_path='checkpoints/non_equal_split/partitions_val.pt')
print('done')
train1, train2 = get_loader_from_dataset(args, train1, True), get_loader_from_dataset(args, train2, False)
val1, val2 = get_loader_from_dataset(args, val1, True), get_loader_from_dataset(args, val2, False)


for batch_ind, batch1 in enumerate(val1):

    image, label = batch1
    image, label = image.cuda(), label.cuda()
    print(label)

    utils.save_image((image + 1) * 0.5, f'{results_folder}/og_img_{batch_ind}.png')
    operation_grad.set_label(label)
    operator.operation.operation_func = operation_grad.get_grads

    gradients = operator.operation.operation_func(image)

    x_new = torch.randn_like(image).requires_grad_(True)
    optimizer = torch.optim.Adam([x_new], lr=0.0001)
    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, 1000)

    for _ in range(1000):
        optimizer.zero_grad()
        new_g = operator.operation.operation_func(x_new)
        loss = mse_loss(new_g, gradients)

        m_loss = loss.sum()
        print(loss)

        m_loss.backward()
        optimizer.step()
        scheduler.step()



    exit()
    print("Start")
    output = operator.operator(label=label, operated_image=gradients)

    utils.save_image((output + 1) * 0.5, f'{results_folder}/new_img_{batch_ind}.png')
    if batch_ind == 0:
        break



